/*
   Copyright (C) 2007,2008,2009,2010 Roger While

   This file is part of OpenCOBOL.

   The OpenCOBOL runtime library is free software: you can redistribute it
   and/or modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.

   OpenCOBOL is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OpenCOBOL.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef COB_LOCAL_H
#define COB_LOCAL_H

/* We use this file to prototype things that should not be
   exported to user space
*/

/* Also OK for icc which defines __GNUC__ */

#if	defined(_WIN32) || defined(__CYGWIN__)
#define COB_HIDDEN	extern
#elif	defined(__GNUC__) && (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 3))
#define COB_HIDDEN	extern __attribute__ ((visibility("hidden")))
#elif	defined(__SUNPRO_C) && (__SUNPRO_C >= 0x550)
#define COB_HIDDEN	extern __hidden
#else
#define COB_HIDDEN	extern
#endif

/* Stacked field level */
#define COB_DEPTH_LEVEL		16

#define COB_FIELD_INIT(x,y,z)	do { \
	field.size = x; \
	field.data = y; \
	field.attr = z; \
	} while (0)

#define COB_ATTR_INIT(v,w,x,y,z)	do { \
	attr.type = v; \
	attr.digits = w; \
	attr.scale = x; \
	attr.flags = y; \
	attr.pic = z; \
	} while (0)

#define COB_GET_SIGN(f)		\
	(COB_FIELD_HAVE_SIGN (f) ? cob_real_get_sign (f) : 0)
#define COB_PUT_SIGN(f,s)	\
	do { if (COB_FIELD_HAVE_SIGN (f)) cob_real_put_sign (f, s); } while (0)

#ifdef	COB_PARAM_CHECK
#define	COB_CHK_PARMS(x,z)	\
	cob_parameter_check (#x, z)
#else
#define	COB_CHK_PARMS(x,z)
#endif

/* Convert a digit (e.g., '0') into an integer (e.g., 0) */
/* RXWRXW
#define COB_D2I(x)		((x) - '0')
*/
#define COB_D2I(x)		((x) & 0x0f)

/* Convert an integer (e.g., 0) into a digit (e.g., '0') */
#define COB_I2D(x)		((x) + '0')

#define	COB_MODULE_PTR		cobglobptr->cob_current_module

COB_HIDDEN void		cob_init_numeric	(struct cob_global_struct *);
COB_HIDDEN void		cob_init_termio		(struct cob_global_struct *);
COB_HIDDEN void		cob_init_fileio		(struct cob_global_struct *);
COB_HIDDEN void		cob_init_call		(struct cob_global_struct *);
COB_HIDDEN void		cob_init_intrinsic	(struct cob_global_struct *);
COB_HIDDEN void		cob_init_strings	(struct cob_global_struct *);
COB_HIDDEN void		cob_init_move		(struct cob_global_struct *);
COB_HIDDEN void		cob_init_screenio	(struct cob_global_struct *);

COB_HIDDEN int		cob_real_get_sign	(cob_field *);
COB_HIDDEN void		cob_real_put_sign	(cob_field *, const int);

COB_HIDDEN void		cob_screen_set_mode	(const unsigned int);
COB_HIDDEN void		cob_set_exception	(const int);
COB_HIDDEN void		cob_field_to_string	(const cob_field *, void *,
						 const size_t);
COB_HIDDEN void		cob_parameter_check	(const char *, const int);
COB_HIDDEN void		cob_runtime_error	(const char *, ...) COB_A_FORMAT12;

#undef	COB_HIDDEN

#endif	/* COB_LOCAL_H */
